// Telos VX phone interface
const vxDTMFsymbols=['0','1','2','3','4','5','6','7','8','9','*','#'];
var vxConnections=[];

/*
TelosVX Protocol

available functions:
connection management
 void phoneVX.OpenConnection(cid, string host);
 void phoneVX.CloseConnection(cid);
 void phoneVX.CloseAllConnections(void);
operations
 void phoneVX.SendMessage(cid, string msg);

available events:
 vxConnected(cid);
 vxDisconnected(cid);
 vxMessage(cid, lwcp_message);
*/

// ------------------------- * Telos VX local protocol logic * -------------------------

function vxCalcNgState(ln)
{
	if(ln.state=='ON_AIR' && (ln.callstate=='DIALING' || ln.callstate=='RINGING_OUT')) return 'ON_AIR_DIALING';
	return ln.state;
}

function vxLineEvent(cid, vx, line_id, props)
{
	if(vx.state!='working') return;
	if(!(line_id in vx.studio.ln_map)) return;
	var ln=vx.studio.ln_map[line_id];
	var chg=false;

	if(('hybrid' in props) && (ln.hybrid!=props.hybrid))
	{
		if(ln.id<0 || (props.hybrid!=0 && props.hybrid<=vx.studio.num_fix))
			log.warning_f("vx%d: line %d hybrid changed from %d to %d", cid, ln.id, ln.hybrid, props.hybrid);
		ln.hybrid=props.hybrid;
		chg=true;
	}
	if('remote' in props)
	{
		var new_remote=(props.remote=='NULL')?"":props.remote;
		if(ln.remote!=new_remote)
		{
			ln.remote=new_remote;
			chg=true;
		}
	}
	if(('state' in props) && ln.state!=props.state)
	{
		ln.state=props.state;
		chg=true;
	}
	if(('callstate' in props) && ln.callstate!=props.callstate)
	{
		ln.callstate=props.callstate;
		chg=true;
	}

/* notify on changes if required */
	if(chg)
	{
		if(ln.id>=0)
		{
			var hyb=ln.hybrid;

			if(hyb!=0) hyb-=vx.studio.num_fix;
			phoneIndiLineState(cid, ln.id, vxCalcNgState(ln), hyb, ln.remote);
		}
		else
			phoneIndiFixedState(cid, -ln.id, vxCalcNgState(ln), ln.remote);
	}
}

function vxStudioEvent(cid, vx, props)
{
	var n, line, sel_n, ln;
	var need_list=false;

	if('name' in props)
	{
		if(vx.st_name!=props.name)
		{
			log.warning_f("vx%d: wrong studio '%s' instead of '%s'", cid, props.name, vx.st_name);
			vx.state='studio';
			phoneVX.SendMessage(cid, "select studio name=\""+lwcpEscape(vx.st_name)+"\"\r\n");
			return;
		}
		if(vx.state=='studio')
		{
			log.info_f("vx%d: connected to studio", cid);
			vx.studio=
			{
				busy:    false, // is busy-all active
				num_ln:  0,     // number of lines
				num_hyb: 0,     // number of hybrids
				num_fix: 0,     // number of fixed hybrids
				num_sel: 0,     // number of selectable lines
				ln_map:  [],    // maps each line to fix/sel, >=0 - sel, <0 - fixed
				sel_map: [],    // maps our selectable line # to VX line #
				fix_map: []     // maps fixed line # to VX line #
			};

			if(vx.sh_name!="")
			{
				vx.state='working';
				need_list=true;
			}
			else
				vx.state='show';
		}
		log.info_f("studio - name %s", props.name);
	}
	if(vx.studio==null || vx.state=='studio') return; /* no studio yet */

	if('show_name' in props)
	{
		if(vx.sh_name!="" && vx.sh_name!=props.show_name)
		{
			log.info_f("vx%d: changing show from '%s' to '%s'", props.show_name, vx.sh_name);
			vx.state='show';
		}
		else if(vx.state=='show')
		{
			vx.state='working';
			need_list=true;
		}
	}
	if(vx.state=='show')
		phoneVX.SendMessage(cid, "select_show studio name=\""+lwcpEscape(vx.sh_name)+"\"\r\n");

	if(('num_lines' in props) && vx.studio.num_ln!=props.num_lines)
	{
		vx.studio.num_ln=props.num_lines;
		log.info_f("vx%d: %d lines", cid, props.num_lines);
		need_list=true;
	}
	if(('num_hyb_fixed' in props) && vx.studio.num_fix!=props.num_hyb_fixed)
	{
		vx.studio.num_fix=props.num_hyb_fixed;
		log.info_f("vx%d: %d fixed", cid, vx.studio.num_fix);
		need_list=true;
	}
	if(('num_hybrids' in props) && vx.studio.num_hyb!=props.num_hybrids)
	{
		vx.studio.num_hyb=props.num_hybrids;
		if(vx.studio.num_hyb<vx.studio.num_fix) log.warning("studio - num_hyb<num_fix");
		log.info_f("vx%d: %d total hybrids (%d selectable)", cid, vx.studio.num_hyb, vx.studio.num_hyb-vx.studio.num_fix);
	}
	if('busy_all' in props) vx.studio.busy=(props.busy_all=='TRUE');

/* process line list last and only if we are ready */
	if(vx.state!='working') return;
	if('line_list' in props)
	{
		vx.studio.ln_map=[];
		vx.studio.sel_map=[];
		vx.studio.fix_map=[];
		sel_n=0;
	/* line list process */
		for(n=0; n<vx.studio.num_ln; n++)
		{
		/* 0: state, 1: callstate, 2: name, 3: local, 4: remote, 5: hybrid, 6: time, 7: comment, 8: direction */
			if(!(n in props.line_list))
			{
				log.warning_f("vx%d: line %d is missing from line_list", cid, n+1);
				line='NULL';
			}
			else
				line=props.line_list[n];
	 /* check line list entry */
			if(line==='NULL' || !(0 in line) || !(5 in line) || !(4 in line))
				line=['NONE','IDLE',"","","",0,0,"",'NONE'];
			if(line[4]=='NULL') line[4]="";
			ln=
			{
				state:     line[0], // line state
				callstate: line[1], // call state
				hybrid:    line[5], // assigned hybrid
				remote:    line[4]  // remote number (caller id)
			};
		/* create fix/sel mapping */
			if(ln.hybrid>=1 && ln.hybrid<=vx.studio.num_fix)
			{
				ln.id= -ln.hybrid; /* fixed line */
				vx.studio.fix_map[ln.hybrid]=n+1;
			}
			else
			{
				ln.id=sel_n; /* selectable line */
				vx.studio.sel_map[sel_n]=n+1;
				sel_n++;
			}
			vx.studio.ln_map[n+1]=ln;
		} 
		vx.studio.num_sel=sel_n;
	/* indicate number and states of lines */
		phoneIndiStudio(cid, vx.studio.num_sel, vx.studio.num_hyb-vx.studio.num_fix, vx.studio.num_fix);
		for(n=0; n<vx.studio.num_sel; n++)
		{
			ln=vx.studio.ln_map[vx.studio.sel_map[n]];
			var hyb=ln.hybrid;
			if(hyb!=0) hyb-=vx.studio.num_fix;
			phoneIndiLineState(cid, n, vxCalcNgState(ln), hyb, ln.remote);
		}
		for(n=1; n<=vx.studio.num_fix; n++)
		{
			if(!(n in vx.studio.fix_map)) continue;
			ln=vx.studio.ln_map[vx.studio.fix_map[n]];
			phoneIndiFixedState(cid, ln.hybrid, vxCalcNgState(ln), ln.remote);
		}
		need_list=false;
	}

	if(need_list)
	{
		vx.studio.ln_map=[];
		vx.studio.sel_map=[];
		log.info_f("vx%d: studio needs new line_list", cid);
		phoneVX.SendMessage(cid, "get studio line_list\r\n");
	}
}

// ------------------------- * Telos VX interface entry point * -------------------------

/*
** vxConnected
**
** called from C code when connection to host is established
*/
function vxConnected(cid)
{
	log.info_f("vxConnected(%d)", cid);
	if(!(cid in vxConnections)) return;

	var vx=vxConnections[cid];
	if(vx.state!='conhost') return;
	
	log.info_f("vx%d: Connected to '%s', logging in", cid, vx.host);
	vx.state='login';
	phoneVX.SendMessage(cid, "login cc user=\""+lwcpEscape(vx.login)+"\",password=\""+lwcpEscape(vx.pass)+"\"\r\n");
}

/*
** vxDisconnected
**
** called from C code when connection to host is broken (by timeout, by device or by our request)
*/
function vxDisconnected(cid)
{
	log.info_f("vxDisconnected(%d)", cid);
	if(!(cid in vxConnections)) return;

	var vx=vxConnections[cid];
	if(vx.state=='idle') return;
	vx.state='conhost';

// TODO: indicate connection loss
	log.info_f("vx%d: Connection lost", cid);
	phoneIndiStudio(cid, 0, 0, 0);
}

/*
** vxDisconnected
**
** called from C code on incoming LWCP message from VX
*/
function vxMessage(cid, obj, op, props, sys)
{
	if(!(cid in vxConnections)) return;

	var vx=vxConnections[cid];
	if(vx.state=='idle') return;

	if(obj==null) return;
/* log all acks, most likely errors, for analisis */
	if(op=='ack' && ('status' in sys) && ('msg' in sys))
	{
		log.warning_f("vx%d: %s %s", cid, sys.status, sys.msg);
		return;
	}
/* process cc messages */
	if(obj.name=='cc' && obj.id==0 && !('sub' in obj))
	{
		if(op=='ack' && ('logged' in props) && vx.state=='login')
		{
			if(props.logged=='TRUE')
			{
				log.info_f("vx%d: Logged in, selecting studio '%s'.", cid, vx.st_name);
				vx.state='studio';
				phoneVX.SendMessage(cid, "select studio name=\""+lwcpEscape(vx.st_name)+"\"\r\n");
			}
			else
				log.warning_f("vx%d: Login failed", cid);
		}
		return;
	}

	if(op!='event' && op!='indi') return; /* care only about events */
/* process studio messages */
	if(obj.name=='studio' && obj.id==0)
	{
		if('sub' in obj)
		{
			if(obj.sub.name=='line' && !('sub' in obj.sub))
			{
			/* process line events */
				vxLineEvent(cid, vx, obj.sub.id, props);
			}
			return;
		}
		vxStudioEvent(cid, vx, props);
	}
}

// ------------------------- * Telos VX protocol commands * -------------------------

/*
** vxConnect
**
** Connection request from scripts
*/
function vxConnect(cid, host, login, pass, studio, show)
{
	log.info_f("vxConnect(%d, %s, %s, %s, %s, %s)", cid, host, login, pass, studio, show);
	if(cid in vxConnections)
	{
		if(vxConnections[cid].state!='idle') phoneVX.CloseConnection(cid);
		delete vxConnections[cid];
	}
	if(host=="" || studio=="" || !phoneVX.OpenConnection(cid, host))
		return;
	vxConnections[cid]=
	{
		state: 'conhost', // {idle, conhost, login, studio, show, working}
		host: host,
		login:login,
		pass: pass,
		st_name: studio,
		sh_name: show,
		studio: null
	};
}

function vxCmndBlockAllToggle(cid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	phoneVX.SendMessage(cid, "busy_all studio state="+(vx.studio.busy?'FALSE':'TRUE')+"\r\n");
}

function vxCmndNext(cid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	phoneVX.SendMessage(cid, "take studio.line\r\n");
}

function vxCmndMute(cid, mute)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	phoneVX.SendMessage(cid, "set studio mute="+(mute?'TRUE':'FALSE')+"\r\n");
}

function vxCmndHybridDrop(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<=0 || hybrid>(vx.studio.num_hyb-vx.studio.num_fix)) return;
	phoneVX.SendMessage(cid, "drop studio hybrid="+(hybrid+vx.studio.num_fix)+"\r\n");
}

function vxCmndHybridHold(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<=0 || hybrid>(vx.studio.num_hyb-vx.studio.num_fix)) return;
	phoneVX.SendMessage(cid, "hold studio hybrid="+(hybrid+vx.studio.num_fix)+"\r\n");
}

function vxCmndLineSeize(cid, line, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	if(hybrid<=0 || hybrid>(vx.studio.num_hyb-vx.studio.num_fix)) return;
	var ln=vx.studio.ln_map[vx.studio.sel_map[line]];
	phoneVX.SendMessage(cid, "seize studio.line#"+vx.studio.sel_map[line]+" hybrid="+(hybrid+vx.studio.num_fix)+"\r\n");
}

function vxCmndLineDrop(cid, line)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	phoneVX.SendMessage(cid, "drop studio.line#"+vx.studio.sel_map[line]+"\r\n");
}

function vxCmndLineHold(cid, line)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	phoneVX.SendMessage(cid, "hold studio.line#"+vx.studio.sel_map[line]+"\r\n");
}

function vxCmndLineCall(cid, line, number, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	if(hybrid<0 || hybrid>(vx.studio.num_hyb-vx.studio.num_fix)) return;
	var msg="call studio.line#"+vx.studio.sel_map[line];
	msg+= " number=\""+lwcpEscape(number)+"\"";
	if(hybrid!=0) msg+= ", hybrid="+(hybrid+vx.studio.num_fix);
	phoneVX.SendMessage(cid, msg+"\r\n");
}

function vxCmndLineDTMF(cid, line, keycode)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	if(!(keycode in vxDTMFsymbols)) return;
	phoneVX.SendMessage(cid, "dtmf studio.line#"+vx.studio.sel_map[line]+" keys='"+vxDTMFsymbols[keycode]+"'\r\n");
}

function vxCmndLineTake(cid, line, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	if(hybrid<0 || hybrid>(vx.studio.num_hyb-vx.studio.num_fix)) return;
	var msg="take studio.line#"+vx.studio.sel_map[line];
	if(hybrid!=0) msg+= " hybrid="+(hybrid+vx.studio.num_fix);
	phoneVX.SendMessage(cid, msg+"\r\n");
}

function vxCmndLineLock(cid, line)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	phoneVX.SendMessage(cid, "lock studio.line#"+vx.studio.sel_map[line]+"\r\n");
}

function vxCmndLineUnlock(cid, line)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(!(line in vx.studio.sel_map)) return;
	phoneVX.SendMessage(cid, "unlock studio.line#"+vx.studio.sel_map[line]+"\r\n");
}

function vxCmndFixedDrop(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<=0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "drop studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedHold(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<=0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "hold studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedTake(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "take studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedSeize(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "seize studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedCall(cid, hybrid, number)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	var msg="call studio.line#"+vx.studio.fix_map[hybrid];
	msg+= " number=\""+lwcpEscape(number)+"\"";
	phoneVX.SendMessage(cid, msg+"\r\n");
}

function vxCmndFixedDTMF(cid, hybrid, keycode)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	if(!(keycode in vxDTMFsymbols)) return;
	phoneVX.SendMessage(cid, "dtmf studio.line#"+vx.studio.fix_map[hybrid]+" keys='"+vxDTMFsymbols[keycode]+"'\r\n");
}

function vxCmndFixedLock(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "lock studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedUnlock(cid, hybrid)
{
	if(!(cid in vxConnections)) return;
	var vx=vxConnections[cid];

	if(vx.state!='working') return;
	if(hybrid<0 || hybrid>vx.studio.num_fix) return;
	if(!(hybrid in vx.studio.fix_map)) return;
	phoneVX.SendMessage(cid, "unlock studio.line#"+vx.studio.fix_map[hybrid]+"\r\n");
}

function vxCmndFixedGetState(cid, hybrid)
{
	var ln={state:'NONE',remote:''};

	if(!(cid in vxConnections)) return ln;
	var vx=vxConnections[cid];

	if(vx.state!='working' || hybrid<0 || hybrid>vx.studio.num_fix || !(hybrid in vx.studio.fix_map)) return ln;
	ln.state=vxCalcNgState(vx.studio.ln_map[vx.studio.fix_map[hybrid]]);
	ln.remote=vx.studio.ln_map[vx.studio.fix_map[hybrid]].remote;
	return ln;
}
